import asyncio
from dataclasses import dataclass
from typing import Dict, Callable, Awaitable
from asyncio.base_events import functools
import websockets
import json
import brotli
from .pack import Pack, RawDanmu, WSConstants


class DanmuClient:
    roomId: int
    ws: websockets.WebSocketClientProtocol
    _handler: Dict[int, Callable[[int, dict | RawDanmu], Awaitable[None]]] ={}
    api = "wss://broadcastlv.chat.bilibili.com:443/sub"
    heartbeat = 30.0

    def __init__(self, roomId: int):
        self._handler[WSConstants.WS_OP_MESSAGE] = self.default_handler
        self._handler[WSConstants.WS_OP_HEARTBEAT_REPLY] = self.default_handler
        self._handler[0] = self.default_handler

        self.roomId = roomId
        self.heartbeat_pack = Pack.pack_string("", WSConstants.WS_OP_HEARTBEAT)

    async def send_auth(self):
        auth_params = {"uid": 1, "roomid": self.roomId, 
                       "protover": 3, "platform": "web", "clientver": "1.7.3"}
        return await self.ws.send(Pack.pack_string(json.dumps(auth_params), WSConstants.WS_OP_USER_AUTHENTICATION))
    
    async def send_heartbeat(self):
        if self.ws.open:
            return await self.ws.send(self.heartbeat_pack)

    async def handle_packs(self, packs: bytes):
        if not packs:
            return None
        header = Pack.unpack_header(packs)
        body = packs[header.headerLength:]
        if header.protocolVersion == 3 and header.operation == WSConstants.WS_OP_MESSAGE:
           packs = brotli.decompress(body)
           for raw in Pack.unpack_string(packs):
                await self.parse_body(raw)
        else: 
            if len(packs) == header.packLength:
                await self.parse_body(RawDanmu(header=header, body=body))

    async def parse_body(self, rawDanmu: RawDanmu):
        opt = rawDanmu.header.operation
        if opt == WSConstants.WS_OP_MESSAGE:
            await self._handler[WSConstants.WS_OP_MESSAGE](self.roomId, json.loads(rawDanmu.body.decode("utf-8")))
        elif opt == WSConstants.WS_OP_HEARTBEAT_REPLY:
            await self._handler[WSConstants.WS_OP_HEARTBEAT_REPLY](self.roomId, rawDanmu)
        else:
            await self._handler[0](self.roomId, rawDanmu)
    
    async def default_handler(self, room_id: int, danmu: dict | RawDanmu) -> None:
        pass

    def on_danmu(self, fn: Callable[[int, dict], Awaitable[None]]):
        self._handler[WSConstants.WS_OP_MESSAGE] = fn
        return fn

    def on_unknown(self, fn: Callable[[int, RawDanmu], Awaitable[None]]):
        self._handler[0] = fn
        return fn

    def on_heartbeat(self, fn: Callable[[int, RawDanmu], Awaitable[None]]):
        # watching_num = struct.unpack("!I", rawDanmu.body)
        self._handler[WSConstants.WS_OP_HEARTBEAT_REPLY] = fn
        return fn

    async def job_send_heartbeat(self):
        try:
            while True:
                await self.send_heartbeat()
                await asyncio.sleep(self.heartbeat)
        except asyncio.CancelledError:
            return

    def run(self, loop: asyncio.AbstractEventLoop):
        async def init_client():
            self.ws = await websockets.connect(self.api, ssl=True)
            await self.send_auth()
            loop.create_task(self.job_send_heartbeat())
        async def receive_packs():
            async for packs in self.ws:
                try:
                    await self.handle_packs(packs)
                except websockets.ConnectionClosed:
                    continue

        async def start_client():
            while True:
                try:
                    await receive_packs()
                except websockets.ConnectionClosed:
                    await self.ws.close()
                    self.ws = await websockets.connect(self.api, ssl=True)
                    await self.send_auth()
                    continue

        def start(*arg):
            loop.create_task(start_client())

        loop.create_task(init_client()).add_done_callback(start)





from typing import Iterator
from struct import Struct
from enum import IntEnum
from dataclasses import dataclass

class WSConstants(IntEnum):
    # 客户端发送心跳值
    WS_OP_HEARTBEAT = 2
    # 服务端返回心跳值
    WS_OP_HEARTBEAT_REPLY = 3
    # 返回消息
    WS_OP_MESSAGE = 5
    # 用户授权加入房间
    WS_OP_USER_AUTHENTICATION = 7
    # 建立连接成功，客户端接收到此信息时需要返回一个心跳包
    WS_OP_CONNECT_SUCCESS = 8
    # Header Length
    WS_PACKAGE_HEADER_TOTAL_LENGTH = 16
    WS_PACKAGE_OFFSET = 0
    WS_HEADER_OFFSET = 4
    WS_VERSION_OFFSET = 6
    WS_OPERATION_OFFSET = 8
    WS_SEQUENCE_OFFSET = 12
    WS_BODY_PROTOCOL_VERSION_NORMAL = 0
    # deflate 压缩版本
    WS_BODY_PROTOCOL_VERSION_DEFLATE = 3
    WS_HEADER_DEFAULT_VERSION = 1
    WS_HEADER_DEFAULT_OPERATION = 1
    WS_HEADER_DEFAULT_SEQUENCE = 1
    WS_AUTH_OK = 0
    WS_AUTH_TOKEN_ERROR = -101

@dataclass
class HeaderInfoItem:
    key: str
    bytes: int
    offset: int
    value: int

class HeaderInfo:
    headerLength = HeaderInfoItem("headerLen", 2, 4, 16)
    protocolVersion = HeaderInfoItem("ver", 2, 6, 1)
    operation = HeaderInfoItem("op", 4, 8, 1)
    sequenceId = HeaderInfoItem("seq", 4, 12, 1)
    itemAmount = 4 
    rawHeaderLength = itemAmount + headerLength.bytes + protocolVersion.bytes + operation.bytes + sequenceId.bytes

@dataclass
class Header:
    packLength: int
    headerLength: int
    protocolVersion: int
    operation: int
    sequenceId: int

@dataclass
class RawDanmu:
    header: Header
    body: bytes 

class Pack:
    struct = Struct(">I2H2I") # I: PackSize; 2H:HeaderLength, ProtocolVersion; 2I:Operation, SequenceId

    @staticmethod
    def pack_header(pack_size: int, opt: int) -> bytes:
        return Pack.struct.pack(pack_size, HeaderInfo.rawHeaderLength, 
                                WSConstants.WS_BODY_PROTOCOL_VERSION_DEFLATE, opt,
                                WSConstants.WS_HEADER_DEFAULT_SEQUENCE)

    @staticmethod
    def unpack_header(header: bytes) -> Header:
        pack_len, header_len, ver, opt, seq = Pack.struct.unpack_from(header)
        if header_len == HeaderInfo.rawHeaderLength:
            return Header(packLength=pack_len, headerLength=header_len,
                          protocolVersion=ver, operation=opt, 
                          sequenceId=seq)
        raise ValueError("Header Length does not equal to Raw Header Length")
    
    @staticmethod
    def pack_string(body: str, opt: int) -> bytes:
        bytes_body = body.encode("utf-8")
        pack_size = len(bytes_body) + HeaderInfo.rawHeaderLength
        header = Pack.pack_header(pack_size, opt)
        return header + bytes_body
    
    @staticmethod
    def unpack_string(packs: bytes) -> Iterator[RawDanmu]:
        pack_offset = 0
        packs_len = len(packs)
        while pack_offset != packs_len:
            header = Pack.unpack_header(packs[pack_offset:pack_offset+HeaderInfo.rawHeaderLength])
            next_pack_offset = pack_offset + header.packLength
            body = packs[pack_offset+HeaderInfo.rawHeaderLength:next_pack_offset]
            yield RawDanmu(header=header, body=body)
            pack_offset = next_pack_offset




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['danmu']

package_data = \
{'': ['*']}

install_requires = \
['brotlipy>=0.7.0,<0.8.0', 'websockets>=10.4,<11.0']

setup_kwargs = {
    'name': 'bili-danmu',
    'version': '0.2.5',
    'description': 'A modern library for receiving danmu from bilibili livestream',
    'long_description': "# bili-danmu\n\nA modern library for receiving danmu from bilibili livestream, with full asynchronous support.\n\nNOTICE: It's a simple implement, so IT DOES NOT INCLUDE PARSEING DANMU FEATURES. You need to parse the danmu dict mannually.\n\n# Installation\n\nJust execute `pip install bili-danmu`\n\n# Example\n\n```python\nimport asyncio\nfrom danmu import DanmuClient\n\nloop = asyncio.new_event_loop()\ndmc = DanmuClient(25512465)\n\n@dmc.on_danmu\nasync def on_danmu(room_id: int, danmu: dict):\n    print(danmu)\n\ndmc.run(loop)\nloop.run_forever()\n```\n\n# Special Thanks\n\nThis project refers to some other projects. Sincerely appreciate their developers.\n\n- [Code4Epoch/Bolaris](https://github.com/Code4Epoch/Bolaris) Give me the infomation of v3 protocol\n\n- [qydysky/bili_danmu](https://github.com/qydysky/bili_danmu) A full list of CMD in danmu pack's body\n\n- [littlecodersh/danmu](https://github.com/littlecodersh/danmu) A very old danmu library. Inspire me to use callback design\n\n- [yjqiang/danmu](https://github.com/yjqiang/danmu) Main reference \n\n- [yangjunlong/barrager.js](https://github.com/yangjunlong/barrager.js) A protocol document and code implement\n",
    'author': 'WorldObservationLog',
    'author_email': 'wolc@duck.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
